import logging
import os
import functools
import requests
from collections import namedtuple
from urllib import urlencode
from urlparse import urlparse
from filelock import FileLock

from pybox.utils import urlopen
from pybox.exceptions import *

logger = logging.getLogger('pybox')

BOX_BASE_URL  = 'https://app.box.com/api/'
BOX_TOKEN_URL = BOX_BASE_URL + 'oauth2/token'

Token = namedtuple('Token', ['access_token', 'refresh_token'])

def retry(attempts=3):
    def wrapper(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            for attempt in range(0, attempts):
                try:
                    return func(*args, **kwargs)
                except BoxAPIException as e:
                    if attempt == attempts - 1:
                        raise
                    else:
                        logger.warn('Unexpected Box error! \n' + str(e))
                        logger.warn('Retrying request')
        return wrapped
    return wrapper

def refreshable(func):
    @functools.wraps(func)
    def wrapped(session, *args, **kwargs):
        try:
            return func(session, *args, **kwargs)
        except BoxAPIException as e:
            if e.error == 'invalid_token':
                logger.info('Box Session expired. Attempting to refresh access token.')
                session.refresh_access_token()
                val = func(session, *args, **kwargs)
                logger.info('Box Session successfully refreshed.')
                return val
            else:
                raise

    return wrapped


class Session(object):
    """
    Box.net session object.
    This class manages Box.net credentials and provides methods
    to get bound urls.
    """
    def __init__(self, client_id, client_secret, access_token=None, refresh_token=None):
        self.client_id     = client_id
        self.client_secret = client_secret
        self.token         = Token(access_token, refresh_token)

    def authorize_url(self, url, state=None):
        """
        :returns authorization_url: url user must visit to authorize specified application.
        """
        data = { 'client_id' : self.client_id,
                 'response_type' : 'code',
                 'redirect_uri' : url }

        if state:
            data['state'] = state

        return 'https://app.box.com/api/oauth2/authorize?%s' % urlencode(data)

    @retry(attempts=3)
    def refresh_access_token(self):
        """
        Using the supplied refresh token, we will try to refresh the
        session's access token.
        :param refresh_token: The refresh token used to refresh the access token.
        :returns refresh_token: The new refresh token that Box generated.
        """
        response = urlopen(requests.Request(url=BOX_TOKEN_URL, data={
           'client_id'     : self.client_id,
           'client_secret' : self.client_secret,
           'refresh_token' : self.token.refresh_token,
           'grant_type'    : 'refresh_token'
        }, method='POST'))
        logger.debug(response)

        self.token = Token(response['access_token'], response['refresh_token'])
        return self.token

    @refreshable
    def authorized_request(self, url, data=None, files={}, headers={},
                                      params={}, method='GET'):
        """
        Wrapper around requests to make authenticated requests to box.net
        """
        logger.debug('Making request to {} with params: {}'.format(url, params))
        # Prevent accidentally leaking credentials
        assert urlparse(url).netloc.endswith('box.com'),\
               'Can only make authorized requests to the Box API'

        access_token = self.token.access_token
        if not access_token:
            raise Exception('No access token attached to this session.\n'
                             + 'Must first authenticate.')

        headers = dict({
            'Authorization' : "Bearer {}".format(access_token)
        }, **headers)

        return urlopen(requests.Request(url=url, data=data, files=files,
                                        headers=headers, method=method,
                                        params=params))



class CLISession(Session):
    """
    Session object that stores session credentials in ~/.pybox for future use.
    """
    def __init__(self, client_id, client_secret, access_token=None, refresh_token=None):
        self._credential_path = os.path.expanduser('~/.pybox')
        self._lock = FileLock(os.path.expanduser('~/.pybox.lock'), timeout=30)
        super(CLISession, self).__init__(client_id, client_secret,
                                         access_token, refresh_token)

    @property
    def token(self):
        if os.path.isfile(self._credential_path):
            with open(self._credential_path, 'r') as f:
                access_token  = f.readline().rstrip()
                refresh_token = f.readline().rstrip()
                return Token(access_token, refresh_token)
        else:
            return None

    @token.setter
    def token(self, token):
        if token.access_token and token.refresh_token:
            with open(self._credential_path, 'w') as f:
                f.write(token.access_token+'\n')
                f.write(token.refresh_token+'\n')

    def refresh_access_token(self, *args, **kwargs):
        with self._lock:
            return super(CLISession, self).refresh_access_token(*args, **kwargs)

